from .node import Node, NodeList
from .helper import get_real_file_path, get_file_type
import json
from .constants import HTML1_TXT, HTML2_TXT

class QGProfiler(object):
    def __init__(self, root_name, file_path):
        self.root_node = Node(root_name, None)
        self.current_node = self.root_node
        self.file_type = get_file_type(file_path)
        self.file_path = get_real_file_path(file_path)

    def push(self, name):
        index = self.current_node.is_child_in_children(name)
        if index == -1:
            new_node = Node(name, self.current_node)
            self.current_node.add_child(new_node)
            self.current_node = new_node
        else:
            self.current_node = self.current_node.get_child(index)
            self.current_node.increment_count()
            self.current_node.modify_time()

    def pop(self):
        if self.root_node != self.current_node:
            self.current_node.increment_value()
            self.current_node.modify_time()
            self.current_node = self.current_node.get_parent()
        else:
            raise ValueError('You have reached root node, try end')

    def end(self):
        if self.root_node == self.current_node:
            self.root_node.increment_value()
            self.root_node.modify_time()
        else:
            raise ValueError('You are not at the root node')

    def generate_file(self):
        def recursive_json_generator(node):
            _dict = {}
            _dict['name'] = node.get_name()
            _dict['value'] = node.get_value()
            _dict['count'] = node.get_count()
            _dict['children'] = [recursive_json_generator(child_node) for child_node in node.get_children()]
            return _dict

        def recursive_xml_generator(node):
            node_name = node.get_name()
            node_value = str(node.get_value())
            node_count = str(node.get_count())
            _xml = '<node '+ 'name="' + node_name + '" value="' + node_value + '" count="' + node_count + '">'
            _xml += ''.join([recursive_xml_generator(child_node) for child_node in node.get_children()]) 
            _xml += '</node>'
            return _xml

        if self.file_type == 'json':
            _json = recursive_json_generator(self.root_node)
            text = json.dumps(_json)
            self.write_file(text)
        elif self.file_type == 'xml':
            text = recursive_xml_generator(self.root_node)
            self.write_file(text)
        elif self.file_type == 'html':
            _json = recursive_json_generator(self.root_node)
            text = json.dumps(_json)
            html_text = HTML1_TXT + text + HTML2_TXT
            self.write_file(html_text)

    def write_file(self, text):
        with open(self.file_path, 'w') as f:
            f.write(text)



from datetime import datetime
from copy import deepcopy

class Node(object):
    def __init__(self, name, parent_node):
        self.__name = name
        self.__modified_time = datetime.utcnow()
        self.__parent_node = parent_node
        self.__children_node = NodeList()
        self.__index_of_child = {}
        self.__value = 0
        self.__count = 1

    def __repr__(self):
        return '<%s.Node object with value %s at %s>' %(self.__name, self.__value, hex(id(self)))

    def get_name(self):
        return self.__name

    def get_value(self):
        return self.__value

    def get_count(self):
        return self.__count

    def get_parent(self):
        return self.__parent_node

    def set_value(self, value):
        self.__value = value

    def set_count(self, count):
        self.__count = count

    def increment_value(self):
        datetime_now = datetime.utcnow()
        time_delta = (datetime_now - self.__modified_time).total_seconds()
        self.__value += time_delta

    def increment_value_by(self, value):
        self.__value += value

    def increment_count(self):
        self.__count += 1

    def increment_count_by(self, count):
        self.__count += count

    def get_children(self):
        return self.__children_node

    def add_child(self, node):
        self.__index_of_child[node.__name] = len(self.__children_node)
        self.__children_node.append(node)

    def is_child_in_children(self, name):
        if name in self.__index_of_child:
            return self.__index_of_child[name]
        else:
            return -1

    def modify_time(self):
        self.__modified_time = datetime.utcnow()

    def get_child(self, index):
        return self.__children_node[index]

    def copy(self):
        return self

    def deepcopy(self):
        return deepcopy(self)


class NodeList(list):
    def item(self, index):
        if 0 <= index < len(self):
            return self[index]

    def __repr__(self):
        return '<NodeList object with %s Nodes at %s>' %(len(self), hex(id(self)))

    length = property(lambda self: len(self),
                      doc="The number of nodes in the NodeList.")

    def copy(self):
        return self

    def deepcopy(self):
        return deepcopy(self)



import os
import re

XML = re.compile("<([/?!]?\w+)|&(#?\w+);|([^<>&'\"=\s]+)|(\s+)|(.)")

def make_folder(folder_path):
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)

def get_file_type(file_path):
    filename = file_path.split('/')[-1]
    if filename.endswith('.json'):
        return 'json'
    elif filename.endswith('.xml'):
        return 'xml'
    elif filename.endswith('.html'):
        return 'html'
    else:
        raise ValueError('filename should either end with .json or .xml' or '.html')

def get_real_file_path(file_path):
    filename = file_path.split('/')[-1]
    folder_path = '/'.join(file_path.split('/')[:-1])
    folder_path = os.path.realpath(os.path.expanduser(folder_path))
    make_folder(folder_path)
    file_path = os.path.join(folder_path, filename)
    return file_path


def xml_scanner(_xml):

    TAG = 1; ENTITY = 2; STRING = 3; WHITESPACE = 4; SEPARATOR = 5

    def gettoken(space=0, scan=XML.scanner(_xml).match):
        try:
            while 1:
                m = scan()
                code = m.lastindex
                text = m.group(code)
                if not space or code != WHITESPACE:
                    return code, text
        except AttributeError:
            raise EOFError

    try:
        while 1:
            code, text = gettoken()
            if code == TAG:
                _type = text[:1]
                if _type == '/':
                    yield 'END', str(text[1:])
                    code, text = gettoken(1)
                    if text != '>':
                        raise SyntaxError, 'not proper end tag'
                else:
                    tag = text
                    attrib = {}
                    while 1:
                        code, text = gettoken(1)
                        if text == '>':
                            yield 'START', str(tag), attrib
                            break
                        if text == '/':
                            yield 'START', str(tag), attrib
                            yield 'END', str(tag)
                            break
                        if text == '?':
                            if _type != text:
                                raise SyntaxError, 'unexpected quotation mark'
                            code, text = gettoken(1)
                            if text != '>':
                                raise SyntaxError, 'expected end tag'
                            break
                        if code == STRING:
                            key = text
                            code, text = gettoken(1)
                            if text != '=':
                                raise SyntaxError, 'expected equal sign'
                            code, quote = gettoken(1)
                            if quote != "'" and quote != '"':
                                raise SyntaxError, 'expected quote'
                            value = []
                            while 1:
                                code, text = gettoken()
                                if text == quote:
                                    break
                                if code == ENTITY:
                                    try:
                                        text = fixentity(text)
                                    except ValueError:
                                        text = text
                                value.append(text)
                            attrib[key] = "".join(value)
    except EOFError:
        pass
    except SyntaxError:
        raise




from node import Node, NodeList
from .qg_profiler import QGProfiler
from .helper import get_real_file_path, get_file_type, xml_scanner
import glob
import json

class QGProfileAggregator(object):
    def __init__(self, in_file_path, out_file_path):
        self.root_node = Node('i_am_root', None)
        self.in_file_path = get_real_file_path(in_file_path)
        get_file_type(out_file_path)
        self.out_file_path = get_real_file_path(out_file_path)

    def add_json(self, _json):
        new_node = self.make_node_from_json(_json, self.root_node)
        new_node_list = NodeList()
        new_node_list.append(new_node)
        self.merge_node_list_to_node(self.root_node, new_node_list)

    def merge_node_list_to_node(self, main_node, node_list):
        for node in node_list:
            index = main_node.is_child_in_children(node.get_name())
            if index == -1:
                main_node.add_child(node)
            else:
                existing_node = main_node.get_child(index)
                existing_node.set_value(node.get_value() + existing_node.get_value())
                existing_node.set_count(node.get_count() + existing_node.get_count())
                self.merge_node_list_to_node(existing_node, node.get_children())

    def make_node_from_json(self, _json, parent_node):
        name = _json['name']
        value = _json['value']
        count = _json['count']
        children = _json['children']
        new_node = Node(name, parent_node)
        new_node.set_value(value)
        new_node.set_count(count)

        for child in children:
            child_node = self.make_node_from_json(child, new_node)
            new_node.add_child(child_node)
        return new_node

    def add_xml(self, _xml):
        current_node = self.root_node
        xml_gen = xml_scanner(_xml)
        for each in xml_gen:
            if each[0] == 'START':
                name = str(each[2]['name'])
                value = float(each[2]['value'])
                count = int(each[2]['count'])
                index = current_node.is_child_in_children(name)
                if index == -1:
                    new_node = Node(name, current_node)
                    new_node.set_value(value)
                    new_node.set_count(count)
                    current_node.add_child(new_node)
                    current_node = new_node
                else:
                    current_node = current_node.get_child(index)
                    current_node.increment_value_by(value)
                    current_node.increment_count_by(count)
            elif each[0] == 'END':
                current_node = current_node.get_parent()

    def generate_file(self):
        for file_path in glob.iglob(self.in_file_path):
            filename = file_path.split('/')[-1]
            if filename.endswith('.json'):
                with open(filename, 'r') as f:
                    raw_json = f.read()
                    _json = json.loads(raw_json)
                    self.add_json(_json)
            elif filename.endswith('.xml'):
                with open(filename, 'r') as f:
                    _xml = f.read()
                    self.add_xml(_xml)

        qg_profiler = QGProfiler('test', self.out_file_path)
        if len(self.root_node.get_children()) == 1:
            qg_profiler.root_node = self.root_node.get_child(0)
        else:
            qg_profiler.root_node = self.root_node
        return qg_profiler.generate_file()



import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='qgprofiler',
    version='0.1.9',
    description='Profiler with time deterministic profiling of functions',
    url='https://github.com/quantumgraph/qgprofiler',
    author='QuantumGraph',
    author_email='contact@quantumgraph.com',
    license='MIT',
    packages=find_packages(),
    long_description=read('README.md'),
    keywords = ['profiler', 'monitoring', 'logging', 'utility'],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Utilities',
        'License :: Freeware',
        'Programming Language :: Python',
    ],
)
